from .ccsv import custom_intersection
from .general import execute_sql, merge_csv, move_frame, sub_run

FUNCTIONS = {
    "sql": execute_sql,
    "pandas_push": move_frame,
    "cat": sub_run,
    "python": sub_run,
    "merge_csv": merge_csv,
    "intersection": custom_intersection,
}



import logging
import sys

from etl.functions import FUNCTIONS
from etl.models import ETL
from etl.utils.utils import file_information, load_yaml

logging.basicConfig()
logging.root.setLevel(logging.NOTSET)
logger = logging.getLogger(__name__)


def run(conf: ETL):
    """Runs the ETL pipeline

    param :conf, ETL: ETL object to run
    """
    for key in conf.order:
        for step in conf.pipeline.steps[key]:
            try:
                for sub_step in step:
                    logger.info("Running %s" % (sub_step))
                    process: dict = step[sub_step]
                    FUNCTIONS[process.get("function", None)](
                        **process, secrets=conf.secrets
                    )
                    if conf.meta_output and process.get("file_output", None):
                        file_information(process.get("file_output"), conf.meta_output)
            except KeyError as ex:
                logger.error("Got key error, skipping...")
                logger.error(ex)
                print(ex)


def main():
    """Main function"""
    args = sys.argv
    if len(args) < 2:
        raise ValueError("Missing pipeline yaml target...")
    conf = load_yaml(args[1])
    conf = ETL.from_dict(conf)
    if conf.secrets:
        conf.secrets = load_yaml(conf.secrets, dynamic=False)
    else:
        conf.secrets = load_yaml(args[2], dynamic=False)
    run(conf)


if __name__ == "__main__":
    """Entrypoint"""
    main()



import os
from typing import Dict, Optional, Tuple, Union

import pandas as pd
from sqlalchemy import create_engine
from sqlalchemy.engine.base import Engine

BASICS = {
    "sqlite": {"delete": "DELETE FROM {table}"},
    "postgres": {"delete": "DELETE FROM {table}"},
}


import logging

logger = logging.getLogger(__name__)


class Database:
    target_host: str = ""
    engine: Optional[Engine] = None
    host_type: str = ""

    def __init__(self, secrets: dict, host: str):
        """Init function"""
        self.secrets = secrets
        self.target_host = host

    def __enter__(self):
        """Entry context, sets engine"""
        host_type, engine = self._get_connection_string()
        self.engine = engine
        self.host_type = host_type
        return self

    def __exit__(self, type, value, traceback):
        """Exit context, disposes engine"""
        if self.engine:
            try:
                self.engine.dispose()
            except AttributeError:
                self.engine.close()

    def execute(
        self,
        sql: Optional[str] = None,
        basic: Optional[str] = None,
        table: Optional[str] = None,
    ) -> Optional[pd.DataFrame]:
        """Execute SQL

        param :optional sql, str: SQL to run--could point to sql file
        param :optional basic, str: Baisc command to run
        param :optional table, str: Table to target

        Returns :Optional[pd.DataFrame]: Dataframe of data
        """
        df = None
        if basic:
            sql: str = BASICS.get(self.host_type, {}).get(basic, None)
            sql = sql.format(table=table)

        if os.path.exists(os.path.dirname(sql)):
            with open(sql, "r") as f:
                sql = " ".join([line.strip() for line in f if line.strip()])

        try:
            df = pd.read_sql(sql, self.engine)
        except Exception as ex:
            logger.error(ex)
            logger.info("continuing...")

        return df

    def _get_connection_string(self) -> Tuple[str, Engine]:
        """Returns correct engine

        Returns :Tuple(str,Engine): host type and engine
        """
        print(self.target_host)
        if self.target_host not in self.secrets.keys():
            raise KeyError(f"Missing {self.target_host} in secrets")

        info = self.secrets.get(self.target_host)
        if "postgres" in info["type"]:
            username, password, host, port, database = (
                info.get("username", None),
                info.get("password", None),
                info.get("host", None),
                info.get("port", None),
                info.get("database", None),
            )
            return "postgres", create_engine(
                f"postgresql://{username}:{password}@{host}:{port}/{database}"
            )
        elif "oracle" in info["type"]:
            import cx_Oracle

            username, password, host, port, service_name, db_alias = (
                info.get("username", None),
                info.get("password", None),
                info.get("host", None),
                info.get("port", None),
                info.get("service_name", None),
                info.get("db_alias", None),
            )
            cx_Oracle.makedsn(host, port, service_name=service_name)
            connection = cx_Oracle.connect(username, password, db_alias)
            return "oracle", connection
        elif "odbc" in info["type"]:
            username, password, host, port, database, driver_path, driver_type = (
                info.get("username", None),
                info.get("password", None),
                info.get("host", None),
                info.get("port", None),
                info.get("database", None),
                info.get("driver_path", None),
                info.get("driver_type", None),
            )

            return "odbc", create_engine(
                f"""{driver_type}://{username}:{password}@{host}/{database}?driver={driver_path}&port={port}&odbc_options='TDS_Version=8.0'"""
            )
        elif "mssql" in info["type"]:
            username, password, port, dsn = (
                info.get("username", None),
                info.get("password", None),
                info.get("port", None),
                info.get("dsn", None),
            )
            return "myssql", create_engine(
                f"mssql+pyodbc://{username}:{password}@{dsn}"
            )
        elif "mysql" in info["type"]:
            username, password, host, port, database = (
                info.get("username", None),
                info.get("password", None),
                info.get("host", None),
                info.get("port", None),
                info.get("database", None),
            )
            return "mysql", create_engine(
                f"mysql://{username}:{password}@{host}:{port}/{database}"
            )
        elif "sqlite" in info["type"]:
            dbfile = info.get("dbfile", None)
            return "sqlite", create_engine(f"sqlite:///{os.path.abspath(dbfile)}")
        else:
            raise NotImplementedError(f"{info['type']} is not supported yet")



import logging
from typing import Optional
from etl.utils.runtime import Database

import logging
import os
import csv

LOG_COMMIT_EXECUTION = {
    "oracle": [
        "CREATE TABLE log_data (id NUMBER GENERATED BY DEFAULT AS IDENTITY PRIMARY KEY, type VARCHAR2(10), log_entry VARCHAR2(1000))",
        "INSERT INTO log_data (log_entry) VALUES (:level_name, :entry)",
    ],
    "postgres": [
        "CREATE TABLE IF NOT EXISTS log_data (id SERIAL PRIMARY KEY, type TEXT, log_entry TEXT)",
        "INSERT INTO log_data (log_entry) VALUES (%s %s)",
    ],
    "sqlite": [
        "CREATE TABLE IF NOT EXISTS log_data (id INTEGER PRIMARY KEY AUTOINCREMENT, type TEXT, log_entry TEXT)",
        "INSERT INTO log_data (type, log_entry) VALUES (?, ?)",
    ],
}


class CSVFileHandler(logging.FileHandler):
    def __init__(self, filename, mode="a", encoding=None, delay=False):
        super().__init__(filename, mode, encoding, delay)
        self.csv_writer = csv.writer(self.stream, quoting=csv.QUOTE_NONNUMERIC)

    def emit(self, record):
        try:
            msg = self.format(record)
            self.csv_writer.writerow([record.levelname, msg])
            self.flush()
        except Exception as e:
            self.handleError(record)


class LogCommit(logging.Logger):
    def __init__(self, name, log_file):
        super().__init__(name)

        if log_file:
            handler = CSVFileHandler(log_file)
            self.addHandler(handler)
            self.log_file = log_file

    def _log(
        self, level, msg, args, exc_info=None, extra=None, stack_info=False, **kwargs
    ):
        super()._log(level, msg, args, exc_info, extra, stack_info, **kwargs)

    def log_commit(self, host, secrets):
        if not self.log_file:
            raise ValueError("No file was sent in")

        with open(self.log_file, "r") as f:
            data = list(csv.reader(f))

        with Database(host, secrets) as db:
            with db.engine.connect() as con:
                _db = LOG_COMMIT_EXECUTION.get(db.host_type)
                if not _db:
                    raise ValueError("DB isn't supported")

                create = _db[0]
                con.execute(create)

                push = _db[1]
                ret = con.execute(push, data)
                return ret



import json
import shutil
import pathlib
from unittest.mock import MagicMock

import pytest

CONTENT = {"foo": "bar"}
CSV_DATA_ONE = [
    "index,name,area",
    "0,test1,rw",
    "1,test2,mb",
    "2,test3,x",
    "3,test4,y",
    "4,test5,z",
]
CSV_DATA_TWO = [
    "index,name,area",
    "0,test1,bb",
    "1,test2,cc",
    "21,test3,rw",
    "32,test4,mb",
    "55,test5,x",
]


@pytest.fixture
def generate_paths(tmp_path: pathlib.Path):
    def _generate(create: bool = False):
        source = tmp_path / "source.json"
        target = tmp_path / "target.json"

        if create:
            source.parent.mkdir(exist_ok=True)
            source.write_text(json.dumps(CONTENT))

        yield source
        yield target

        shutil.rmtree(source, ignore_errors=True)

    return _generate


@pytest.fixture
def generate_intersection(tmp_path: pathlib.Path):
    def _generate():
        source = tmp_path / "csv1.csv"
        target = tmp_path / "csv2.csv"

        source.parent.mkdir(exist_ok=True)
        target.parent.mkdir(exist_ok=True)

        # create some data!
        source.write_text("\n".join(CSV_DATA_ONE))
        target.write_text("\n".join(CSV_DATA_TWO))

        yield source
        yield target
        yield tmp_path

        shutil.rmtree(source, ignore_errors=True)
        shutil.rmtree(target, ignore_errors=True)

    return _generate



import csv
import logging
import subprocess
from pathlib import Path

logger = logging.getLogger(__name__)


def read_csv_generator(filename):
    with open(filename, "r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            yield row


def intersection():
    csv1 = read_csv_generator("csv1.csv")
    csv1_dict = {row["id"]: row for row in csv1}

    intersections = []
    for row in read_csv_generator("csv2.csv"):
        if row["id"] in csv1_dict:
            intersections.append(row)
    return intersections


def custom_intersection(
    csv_one: str,
    csv_two: str,
    out_dir: str,
    column: int,
    inverse: bool = False,
    **_
) -> None:
    csv_one_column = str(Path(out_dir) / f"csv1_col{column}.txt")
    csv_two_column = str(Path(out_dir) / f"csv2_col{column}.txt")
    intersections = str(Path(out_dir) / "intersections.txt")
    intersections_one = str(Path(out_dir) / "intersections_one.txt")
    intersections_two = str(Path(out_dir) / "intersections_two.txt")
    final_intersections = str(Path(out_dir) / "final_intersections.txt")

    flag = "-Ff" if not inverse else "-vFf"

    pipeline = {
        f"Cutting column {column} from {csv_one}": {"cmd": ["cut", "-d,", f"-f{column}", csv_one], "output": csv_one_column},
        f"Cutting column {column} from {csv_two}": {"cmd": ["cut", "-d,", f"-f{column}", csv_two], "output": csv_two_column},
        "Getting inital intersection": {
            "cmd": ["grep", "-Fxf", csv_one_column, csv_two_column],
            "output": intersections,
        },
        f"Saving all intersection for {csv_one}": {
            "cmd": ["grep", flag, intersections, csv_one],
            "output": intersections_one,
        },
        f"Saving all intersections for {csv_two}": {
            "cmd": ["grep", flag, intersections, csv_two],
            "output": intersections_two,
        },
        f"Deleting header row from {intersections_two}": {
            "cmd": ["sed", "-i", "1d", intersections_two]
        },
        "Running paste": {
            #"cmd": ["paste", intersections_one, intersections_two],
            "cmd": ["cat", intersections_one, intersections_two],
            "output": final_intersections
        },
    }

    for key, value in pipeline.items():
        logger.info("%s" % key)
        cmd, out = (value.get("cmd"), value.get("output"))
        if out:
            with open(out, "w") as outfile:
                subprocess.run(cmd, stdout=outfile)
        else:
            subprocess.run(cmd)



import logging
import os
import subprocess
from pathlib import Path

import pandas as pd

from etl.utils.database import Database
from etl.utils.exceptions import NotSupported

logger = logging.getLogger(__name__)


def execute_sql(
    connection: str, file_output: str, sql: str, secrets: dict, output: str = "csv", **_
) -> None:
    """Execute SQL

    param :title, str: Title about to be ran
    param :step, Step: Step object
    param :secrets, dict: Connection information
    param :output, str, 'csv': Output file type
    """
    df = None

    with Database(secrets, connection) as database:
        df = database.execute(sql)

    save_frame(df, file_output, output)


def save_frame(df: pd.DataFrame, filename: str, output_type: str = "csv", **_):
    if df is not None:
        p = Path(filename)
        p.parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"Saving {filename}")

        if output_type == "csv":
            df.to_csv(filename)
        elif output_type == "hdfs":
            df.to_hdf(filename)
        elif output_type == "parquet":
            df.to_parquet(filename)
        else:
            raise NotSupported(f"{output_type}")


def merge_csv(csv_one: str, csv_two: str, column: str, out: str, **_) -> None:
    """Merges based on the item"""

    one: pd.DataFrame = pd.read_csv(csv_one)
    two: pd.DataFrame = pd.read_csv(csv_two)
    out: pd.DataFrame = pd.merge(one, two, on=column)
    save_frame(out, out, "csv")

    # chunk_container: pd.DataFrame = pd.read_csv("filename", chunksize=5000)
    # for chunk in chunk_container:
    #     chunk.to_csv("output", mode="a", index=False)


def move_frame(
    connection: str,
    target_tables: str,
    target_connection: str,
    schema: str,
    secrets: dict,
    sql: str,
    **_,
) -> None:
    """Moves data from one table to other tables

    param :title, str: Title about to be ran
    param :step, Step: Step object
    param :secrets, dict: Connection information
    """
    with Database(secrets, connection) as source:
        with Database(secrets, target_connection) as target:
            for df_chunk in pd.read_sql_query(sql, source.engine, chunksize=10):
                for table in target_tables:
                    inital = True
                    if inital:
                        target.execute(basic="delete", table=table)
                        inital = False

                    logger.info("Moving data")
                    df_chunk.to_sql(
                        table,
                        target.engine,
                        schema=schema,
                        method="multi",
                        index=False,
                        if_exists="append",
                    )


def sub_run(app: str, io: str) -> None:
    """Runs command on a subprocess

    param :step, Step: Step object
    param :output, str, 'csv': Output file type
    """
    logger.info(subprocess.call([app, io]))



import calendar
import json
import os
import subprocess
from datetime import datetime
from pathlib import Path

import yaml
from dynamic_yaml import dump as ddump
from dynamic_yaml import load as dload

from .exceptions import NotSupported

today = datetime.today()
RUNTIME_DEFAULTS = {
    "FIRST_DAY_MONTH": today.replace(day=1).day,
    "LAST_DAY_MONTH": calendar.monthrange(today.year, today.month)[1],
    "YEAR": today.year,
    "MONTH": today.month,
    "DAY": today.day,
}

OUTPUT_FOLDER = f"output-meta/{today.day}-{today.month}-{today.year}"

refresh = lambda data: yaml.safe_load(ddump(dload(ddump(data))))


def load_yaml(name: str, dynamic=True) -> dict:
    """Loads yaml and injects dynamic runtime if dynamic is true

    param :name, str: File to load
    param :dyanmic, bool, true: Use dynamic replacement
    Returns :dict: Dictionary with the dynamic variables updated
    """
    with open(name) as f:
        config = yaml.safe_load(f)

    if dynamic:
        config.update(RUNTIME_DEFAULTS)
        config = refresh(config)
    return config


def file_information(file_name: str, output_location: str) -> None:
    """Saves file metadata

    param :file_name, str: Output location of target file
    param :output-location, str: Output location
    """
    fn, ft = os.path.splitext(os.path.basename(file_name))
    exists = os.path.exists(file_name)
    information = {
        "Size": os.path.getsize(file_name) if exists else None,
        "Date": os.path.getmtime(file_name) if exists else None,
        "Type": str(subprocess.check_output(["file", "-i", "-s", file_name]))
        if exists
        else None,
        "Columns": None,
        "Count": None,
    }

    if ft == ".parquet" and exists:
        raise NotImplementedError("HDF5 isn't implemented yet")
    elif ft == ".csv" and exists:
        information["Count"] = str(subprocess.check_output(["wc", "-l", file_name]))
        information["Columns"] = str(
            subprocess.check_output(["head", "-n1", file_name])
        )
    elif ft == ".hdf5" and exists:
        raise NotImplementedError("HDF5 isn't implemented yet")

    print("saving metadata...")
    out = Path(f"{output_location}/{OUTPUT_FOLDER}/{fn}{ft}.info")
    out.parent.mkdir(parents=True, exist_ok=True)
    with open(out, "w") as f:
        json.dump(information, f, indent=4)



import hashlib
import json
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from pprint import pprint
from typing import List, Optional, Tuple, Union

import yaml

from .database import Database


@dataclass
class Archive:
    """Archive Object"""

    date: datetime
    payload: str
    key: str
    field_1: Optional[Union[str, dict]]
    field_2: Optional[Union[str, dict]]

    def to_json(self) -> dict:
        """Converts class to json"""
        return {"key": self.key, "date": self.date, "payload": self.payload}


class Runtime:
    """Runtime Object"""

    source_data: dict = {}
    target_data: dict = {}
    archives: List = []
    source: Union[str, Path] = ""
    target: Optional[Union[str, Path]] = None
    database: Optional[Database] = None
    database_information: Optional[Tuple[str, dict]] = None

    def __init__(
        self, source: str, target: Optional[str] = None, database: bool = False
    ):
        """Init function

        :param source: Source target host or target file
        :param target: Optional, output file
        """
        self.source = Path(source)
        self.target = target
        self.source_data = {}
        self.target_data = {}
        self.archives = []

        if database:
            with open(Path(target)) as f:
                secrets = yaml.load(f)
            self.database_information = (source, secrets)

        else:
            if self.source.exists():
                with open(self.source, "r") as source_file:
                    self.source_data = json.load(source_file)

            if not self.target and not self.source.exists():
                self.target = self.source

            if not self.target:
                today = datetime.today()
                file_name = f"{today.day}-{today.month}-{today.hour}-{today.minute}-{today.second}.json"
                self.target = Path(f"{self.source.parent}/{file_name}")

            self.target = Path(self.target)

    def archive(
        self,
        location: str,
        payload: str,
        field_1: Optional[dict] = None,
        field_2: Optional[dict] = None,
        sql: Optional[str] = None,
    ) -> None:
        """Archives the file"""

        arch = Archive(
            str(datetime.now()),
            payload,
            hashlib.md5(payload.encode("utf-8")).hexdigest(),
            field_1,
            field_2,
        )
        self.archives.append(arch)

        if self.database_information:
            if not sql:
                raise ValueError("Missing sql statement")
            if "INSERT" not in sql:
                raise ValueError("Missing 'INSERT' statement")

            h, s = self.database_information
            with Database(s, h) as db:
                with db.engine.connect() as con:
                    con.execute(sql, payload)

        else:
            location = Path(location)
            with open(location, "w") as archive_output:
                json.dump(arch.to_json(), archive_output)

    def print_archive(self, target: int = 0) -> None:
        """Prints the target"""
        pprint(self.archives[-target:], indent=4)

    def save(self) -> None:
        """Saves the runtime file"""
        with open(self.target, "w") as target_file:
            write_dictionary = (
                {**self.source_data, **self.target_data}
                if self.source_data
                else self.target_data
            )
            json.dump(write_dictionary, target_file, indent=4)

    def __enter__(self):
        """Entry context, loads the file"""
        return self

    def __exit__(self, type, value, traceback) -> None:
        """Exit context, saves target_data"""
        self.save()



import logging
import os
from etl.utils.log import LogCommit
from etl.utils.database import Database
import pytest
import csv

import logging


@pytest.fixture(scope="module")
def logger(tmpdir_factory):
    # Create a temporary log file path
    log_file = tmpdir_factory.mktemp("logs").join("test_log.txt")

    # Create an instance of the InjectingLogger with file streaming
    logger = LogCommit("Log_Commit", log_file)
    logger.setLevel(logging.DEBUG)
    return logger


@pytest.fixture(scope="module")
def database_url():
    # Use an in-memory SQLite database for testing
    return "sqlite:///:memory:"


def test_log_messages(logger):
    # Log some messages
    logger.info("This is an info message")
    logger.warning("This is a warning message")
    logger.error("This is an error message")

    # Ensure log messages are stored in log_data

    with open(logger.log_file, "r") as f:
        data = list(csv.reader(f))

    assert len(data) == 3
    print(data[0][1])
    assert data[0][1] == "This is an info message"
    assert data[1][1] == "This is a warning message"
    assert data[2][1] == "This is an error message"


def test_log_commit(logger, database_url):
    # Log some messages
    logger.info("Log message 1")
    logger.warning("Log message 2")
    logger.error("Log message 3")

    secrets = {"sqlite": {"type": "sqlite", "dbfile": ":memory:"}}
    inserts = logger.log_commit(secrets, "sqlite")
    # assert inserts.rowcount == 3

    secrets = {"sqlite": {"type": "sqlite", "dbfile": ":memory:"}}
    expected_messages = ["Log message 1", "Log message 2", "Log message 3"]
    with Database(secrets, "sqlite") as db:
        ret = db.execute("SELECT * FROM log_data")
        print(list(ret["log_entry"]))
        for item1, item2 in zip(expected_messages, list(ret["log_entry"])):
            assert item1 == item2


# Run the tests
if __name__ == "__main__":
    pytest.main()



from dataclasses import dataclass, fields
from typing import Dict, List, Optional


@dataclass
class Pipeline:
    out_type: Optional[str] = None
    steps: Dict[str, dict] = None

    @classmethod
    def from_dict(cls, data: dict):
        class_fields = {f.name for f in fields(cls)}
        # retain order pls
        ks = set(data.keys()) - class_fields
        steps_ = {k: [] for k in ks}
        for key in steps_:
            steps_[key] = data[key]
        data["steps"] = steps_
        return cls(**{k: v for k, v in data.items() if k in class_fields})


@dataclass
class ETL:
    cycle: str = "daily"
    order: List[str] = None
    pipeline: List[dict] = None
    secrets: Optional[dict] = None
    meta_output: Optional[str] = None

    def __post_init__(self):
        self.pipeline = Pipeline.from_dict(self.pipeline)

    @classmethod
    def from_dict(cls, data: dict):
        data.pop("parameters")
        [
            data.pop(key)
            for key in ["FIRST_DAY_MONTH", "LAST_DAY_MONTH", "YEAR", "MONTH", "DAY"]
        ]

        class_fields = {f.name for f in fields(cls)}
        return cls(**{k: v for k, v in data.items() if k in class_fields})



import os
from setuptools import find_packages, setup

setup(
    name="generic-etl",
    entry_points={"console_scripts": ["etl=etl.etl:main"]},
    version=os.environ.get("RELEASE"),
    packages=find_packages(),
    install_requires=[
        "dynamic-yaml==1.3.3",
        "greenlet==1.1.3.post0",
        "numpy==1.23.4",
        "pandas==1.5.1",
        "psycopg2-binary==2.9.4",
        "pyaml==21.10.1",
        "python-dateutil==2.8.2",
        "pytz==2022.5",
        "PyYAML==6.0",
        "six==1.16.0",
        "SQLAlchemy==1.4.42",
        "cx-Oracle==8.3.0",
    ],
    author="Christian Decker",
)
